from __future__ import annotations

from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from .blocks import AttentiveStatsPool, DepthwiseConv1d


class Token2SV(nn.Module):
    def __init__(
        self, Q: int, V: int, d: int = 192, out_dim: int = 256, dropout: float = 0.05
    ):
        super().__init__()
        self.Q, self.V = int(Q), int(V)
        self.emb = nn.Embedding(self.Q * self.V, d)

        initial_weights = torch.linspace(1.0, 0.1, self.Q)
        self.cb_weights = nn.Parameter(initial_weights)

        self.enc = nn.Sequential(
            DepthwiseConv1d(d, 7, causal=False),
            nn.GELU(),
            nn.Dropout(dropout),
            DepthwiseConv1d(d, 7, causal=False),
            nn.GELU(),
        )

        self.pool = AttentiveStatsPool(d)
        self.proj = nn.Linear(2 * d, out_dim)

    def _get_mixed_embedding(self, embed_btqd: torch.Tensor) -> torch.Tensor:
        w = F.softmax(self.cb_weights, dim=0).view(1, 1, self.Q, 1)
        return (embed_btqd * w).sum(dim=2)

    def forward(
        self, tokens_btq: torch.Tensor, lengths: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        B, T, Q = tokens_btq.shape
        q_idx = torch.arange(Q, device=tokens_btq.device, dtype=torch.long).view(
            1, 1, Q
        )
        idx = q_idx * self.V + tokens_btq.long()
        raw_emb = self.emb(idx)

        if self.training:
            keep_prob = 0.95
            mask = torch.rand(B, T, device=tokens_btq.device) < keep_prob
            bad = mask.sum(dim=1) == 0
            if bad.any():
                bad_idx = bad.nonzero(as_tuple=False).squeeze(1)
                rand_pos = torch.randint(
                    0, T, (bad_idx.numel(),), device=tokens_btq.device
                )
                mask[bad_idx, rand_pos] = True
            raw_emb = raw_emb * mask.float().unsqueeze(-1).unsqueeze(-1)

        x = self._get_mixed_embedding(raw_emb)
        h = self.enc(x)
        pooled = self.pool(h, lengths=lengths)
        e = self.proj(pooled)
        return F.normalize(e, dim=-1, eps=1e-6)


class SpeakerFiLM(nn.Module):
    def __init__(self, d_model: int, sv_dim: int):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(sv_dim, d_model),
            nn.GELU(),
            nn.Linear(d_model, 2 * d_model),
        )
        self.norm = nn.LayerNorm(d_model)
        nn.init.zeros_(self.mlp[-1].weight)
        nn.init.zeros_(self.mlp[-1].bias)

    def forward(
        self, base_bt_d: torch.Tensor, spk_b_d: torch.Tensor, strength: float = 1.0
    ) -> torch.Tensor:
        B, T, D = base_bt_d.shape
        film = self.mlp(spk_b_d)
        gamma, beta = film.chunk(2, dim=-1)
        gamma = gamma.unsqueeze(1).expand(B, T, D)
        beta = beta.unsqueeze(1).expand(B, T, D)
        x = self.norm(base_bt_d)
        return x * (1 + strength * torch.tanh(gamma)) + strength * torch.tanh(beta)



from __future__ import annotations

import torch
import torch.nn as nn
import torch.nn.functional as F


class GLU(nn.Module):
    def __init__(self, d: int):
        super().__init__()
        self.pro = nn.Linear(d, 2 * d)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a, b = self.pro(x).chunk(2, dim=-1)
        return a * torch.sigmoid(b)


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x32 = x.float()
        var = x32.pow(2).mean(dim=-1, keepdim=True)
        y32 = x32 * torch.rsqrt(var + self.eps)
        y32 = y32 * self.weight.float()
        return y32.to(dtype=x.dtype)


class DepthwiseConv1d(nn.Module):
    def __init__(
        self, d: int, kernel_size: int = 7, causal: bool = False, dilation: int = 1
    ):
        super().__init__()
        self.causal = causal
        self.dilation = dilation
        self.kernel_size = kernel_size
        self.dw = nn.Conv1d(d, d, kernel_size, groups=d, padding=0, dilation=dilation)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        xt = x.transpose(1, 2)
        if self.causal:
            pad_left = (self.kernel_size - 1) * self.dilation
            xt = F.pad(xt, (pad_left, 0))
        else:
            total = (self.kernel_size - 1) * self.dilation
            left = total // 2
            right = total - left
            xt = F.pad(xt, (left, right))
        y = self.dw(xt)
        return y.transpose(1, 2)


class SSMLiteBlock(nn.Module):
    def __init__(
        self,
        d: int,
        dropout: float = 0.05,
        causal: bool = False,
        kernel_size: int = 7,
        dilation: int = 1,
    ):
        super().__init__()
        self.norm = RMSNorm(d)
        self.glu = GLU(d)
        self.dw = DepthwiseConv1d(
            d, kernel_size=kernel_size, causal=causal, dilation=dilation
        )
        self.ff = nn.Sequential(
            RMSNorm(d),
            nn.Linear(d, 4 * d),
            nn.GELU(),
            nn.Linear(4 * d, d),
        )
        self.drop = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        h = self.glu(self.norm(x))
        h = self.dw(h)
        x = x + self.drop(h)
        x = x + self.drop(self.ff(x))
        return x


class AttentiveStatsPool(nn.Module):
    def __init__(self, d: int):
        super().__init__()
        self.attn = nn.Sequential(
            nn.Linear(d, d),
            nn.Tanh(),
            nn.Linear(d, 1),
        )

    def forward(
        self, h: torch.Tensor, lengths: torch.Tensor | None = None
    ) -> torch.Tensor:
        B, T, D = h.shape
        logits = self.attn(h).squeeze(-1)

        if lengths is not None:
            mask = torch.arange(T, device=h.device)[None, :] < lengths[:, None]
            logits = logits.masked_fill(~mask, -1e9)

        w = torch.softmax(logits, dim=1).unsqueeze(-1)
        mu = (h * w).sum(dim=1)
        var = (w * (h - mu.unsqueeze(1)).pow(2)).sum(dim=1).clamp_min(1e-6)
        std = torch.sqrt(var)
        return torch.cat([mu, std], dim=-1)



from __future__ import annotations

import argparse
import time

import torch
from tqdm.auto import tqdm

from .audio import save_audio
from .constants import TARGET_SR
from .model import SoproTTS


def main() -> None:
    ap = argparse.ArgumentParser(description="SoproTTS cli inference")

    ap.add_argument("--repo_id", type=str, default="samuel-vitorino/sopro")
    ap.add_argument(
        "--revision", type=str, default=None, help="Optional git revision/branch/tag"
    )
    ap.add_argument("--cache_dir", type=str, default=None, help="Optional HF cache dir")
    ap.add_argument("--hf_token", type=str, default=None, help="HF token")

    ap.add_argument("--text", type=str, required=True)
    ap.add_argument("--ref_audio", type=str, default=None)
    ap.add_argument("--ref_tokens", type=str, default=None, help="Path to .npy [T,Q]")
    ap.add_argument("--out", type=str, required=True)

    ap.add_argument("--max_frames", type=int, default=400)

    ap.add_argument("--top_p", type=float, default=0.9)
    ap.add_argument("--temperature", type=float, default=1.05)
    ap.add_argument("--no_anti_loop", action="store_true")

    ap.add_argument("--no_prefix", action="store_true")
    ap.add_argument("--prefix_sec", type=float, default=None)
    ap.add_argument("--style_strength", type=float, default=None)
    ap.add_argument("--ref_seconds", type=float, default=None)

    ap.add_argument("--seed", type=int, default=None, help="Random seed for sampling")

    ap.add_argument(
        "--no_stop_head", action="store_true", help="Disable stop head early stopping"
    )
    ap.add_argument(
        "--stop_patience", type=int, default=None, help="Override cfg.stop_patience"
    )
    ap.add_argument(
        "--stop_threshold", type=float, default=None, help="Override cfg.stop_threshold"
    )

    ap.add_argument(
        "--device",
        type=str,
        default=None,
        choices=["cpu", "cuda", "mps"],
        help="Device to run on (default: cuda if available else cpu)",
    )

    ap.add_argument("--quiet", action="store_true")

    args = ap.parse_args()

    default_device = "cuda" if torch.cuda.is_available() else "cpu"
    device = args.device or default_device

    if device == "cuda" and not torch.cuda.is_available():
        raise SystemExit("Error: --device cuda requested but CUDA is not available.")
    if device == "mps" and not (
        hasattr(torch.backends, "mps") and torch.backends.mps.is_available()
    ):
        raise SystemExit("Error: --device mps requested but MPS is not available.")

    if args.seed is not None:
        torch.manual_seed(args.seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(args.seed)

    t0 = time.perf_counter()
    tts = SoproTTS.from_pretrained(
        args.repo_id,
        revision=args.revision,
        cache_dir=args.cache_dir,
        token=args.hf_token,
        device=device,
    )
    t1 = time.perf_counter()
    if not args.quiet:
        print(f"[Load] {t1 - t0:.2f}s")

    cfg = tts.cfg

    ref_tokens_tq = None
    if args.ref_tokens is not None:
        import numpy as np

        arr = np.load(args.ref_tokens)
        ref_tokens_tq = torch.from_numpy(arr).long()

    with torch.inference_mode():
        text_ids = tts.encode_text(args.text)
        ref = tts.encode_reference(
            ref_audio_path=args.ref_audio,
            ref_tokens_tq=ref_tokens_tq,
            ref_seconds=args.ref_seconds,
        )

        prep = tts.model.prepare_conditioning(
            text_ids,
            ref,
            max_frames=args.max_frames,
            device=tts.device,
            style_strength=float(
                args.style_strength
                if args.style_strength is not None
                else cfg.style_strength
            ),
        )

        t_start = time.perf_counter()

        hist_A: list[int] = []
        pbar = tqdm(
            total=args.max_frames,
            desc="AR sampling",
            unit="frame",
            disable=args.quiet,
        )

        for _t, rvq1, p_stop in tts.model.ar_stream(
            prep,
            max_frames=args.max_frames,
            top_p=args.top_p,
            temperature=args.temperature,
            anti_loop=(not args.no_anti_loop),
            use_prefix=(not args.no_prefix),
            prefix_sec_fixed=args.prefix_sec,
            use_stop_head=(False if args.no_stop_head else None),
            stop_patience=args.stop_patience,
            stop_threshold=args.stop_threshold,
        ):
            hist_A.append(int(rvq1))
            pbar.update(1)
            if p_stop is None:
                pbar.set_postfix(p_stop="off")
            else:
                pbar.set_postfix(p_stop=f"{float(p_stop):.2f}")

        pbar.n = len(hist_A)
        pbar.close()

        t_after_sampling = time.perf_counter()

        T = len(hist_A)
        if T == 0:
            save_audio(args.out, torch.zeros(1, 0), sr=TARGET_SR)
            t_end = time.perf_counter()
            if not args.quiet:
                print(
                    f"[Timing] sampling={t_after_sampling - t_start:.2f}s, "
                    f"postproc+decode+save={t_end - t_after_sampling:.2f}s, "
                    f"total={t_end - t_start:.2f}s"
                )
                print(f"[Done] Wrote {args.out}")
            return

        tokens_A = torch.tensor(hist_A, device=tts.device, dtype=torch.long).unsqueeze(0)
        cond_seq = prep["cond_all"][:, :T, :]
        tokens_1xTQ = tts.model.nar_refine(cond_seq, tokens_A)
        tokens_tq = tokens_1xTQ.squeeze(0)

        wav = tts.codec.decode_full(tokens_tq)
        save_audio(args.out, wav, sr=TARGET_SR)

    t_end = time.perf_counter()
    if not args.quiet:
        print(
            f"[Timing] sampling={t_after_sampling - t_start:.2f}s, "
            f"postproc+decode+save={t_end - t_after_sampling:.2f}s, "
            f"total={t_end - t_start:.2f}s"
        )
        print(f"[Done] Wrote {args.out}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import time
from dataclasses import dataclass
from typing import Iterator, List, Optional, Tuple

import torch

from .codec.mimi import MimiDecodeState, MimiStreamDecoder
from .model import SoproTTS, SoproTTSModel


@dataclass
class StreamConfig:
    chunk_frames: int = 16
    nar_context_frames: Optional[int] = None
    cond_chunk_size: int = 32


class SoproTTSStreamer:
    def __init__(self, tts: SoproTTS, cfg: Optional[StreamConfig] = None):
        self.tts = tts
        self.cfg = cfg or StreamConfig()
        self.mimi_stream = MimiStreamDecoder(tts.codec)

    @torch.inference_mode()
    def stream(
        self,
        text: str,
        *,
        ref_audio_path: Optional[str] = None,
        ref_tokens_tq: Optional[torch.Tensor] = None,
        max_frames: int = 400,
        top_p: float = 0.9,
        temperature: float = 1.05,
        anti_loop: bool = True,
        use_prefix: bool = True,
        prefix_sec_fixed: Optional[float] = None,
        style_strength: Optional[float] = None,
        ref_seconds: Optional[float] = None,
        chunk_frames: Optional[int] = None,
        nar_context_frames: Optional[int] = None,
        cond_chunk_size: Optional[int] = None,
        use_stop_head: Optional[bool] = None,
        stop_patience: Optional[int] = None,
        stop_threshold: Optional[float] = None,
        min_gen_frames: Optional[int] = None,
    ) -> Iterator[torch.Tensor]:
        model: SoproTTSModel = self.tts.model
        device = self.tts.device

        text_ids = self.tts.encode_text(text)
        ref = self.tts.encode_reference(
            ref_audio_path=ref_audio_path,
            ref_tokens_tq=ref_tokens_tq,
            ref_seconds=ref_seconds,
        )

        prep = model.prepare_conditioning_lazy(
            text_ids,
            ref,
            max_frames=max_frames,
            device=device,
            style_strength=float(
                style_strength
                if style_strength is not None
                else self.tts.cfg.style_strength
            ),
        )

        cf = int(chunk_frames if chunk_frames is not None else self.cfg.chunk_frames)
        cond_cs = int(
            cond_chunk_size if cond_chunk_size is not None else self.cfg.cond_chunk_size
        )

        nar_ctx = (
            nar_context_frames
            if nar_context_frames is not None
            else self.cfg.nar_context_frames
        )
        if nar_ctx is None:
            nar_ctx = int(model.rf_nar())
        nar_ctx = int(nar_ctx)

        hist_A: List[int] = []

        frames_emitted = 0

        mimi_state = MimiDecodeState()

        def refine_and_emit(end: int) -> Optional[torch.Tensor]:
            nonlocal frames_emitted, mimi_state

            new_start = frames_emitted
            if end <= new_start:
                return None

            win_start = max(0, new_start - nar_ctx)
            win_end = end

            cond_win = prep["cond_all"][:, win_start:win_end, :]
            tokens_A_win = torch.as_tensor(
                hist_A[win_start:win_end], device=device, dtype=torch.long
            ).unsqueeze(0)

            tokens_win_tq = model.nar_refine(cond_win, tokens_A_win).squeeze(0)

            tail_i = new_start - win_start
            emit_tokens = tokens_win_tq[tail_i:, :]

            wav_chunk, mimi_state = self.mimi_stream.decode_step(
                emit_tokens, mimi_state
            )
            frames_emitted = end
            return wav_chunk if wav_chunk.numel() > 0 else None

        for _t, rvq1_id, _p_stop in model.ar_stream(
            prep,
            max_frames=max_frames,
            top_p=top_p,
            temperature=temperature,
            anti_loop=anti_loop,
            use_prefix=use_prefix,
            prefix_sec_fixed=prefix_sec_fixed,
            cond_chunk_size=cond_cs,
            use_stop_head=use_stop_head,
            stop_patience=stop_patience,
            stop_threshold=stop_threshold,
            min_gen_frames=min_gen_frames,
        ):
            hist_A.append(int(rvq1_id))
            T = len(hist_A)

            is_boundary = (T % cf) == 0
            if not is_boundary and T < max_frames:
                continue

            wav = refine_and_emit(T)
            if wav is not None:
                yield wav

        T_final = len(hist_A)
        if frames_emitted < T_final:
            wav = refine_and_emit(T_final)
            if wav is not None:
                yield wav

@torch.inference_mode()
def stream(
    tts: SoproTTS,
    text: str,
    *,
    ref_audio_path: Optional[str] = None,
    ref_tokens_tq: Optional[torch.Tensor] = None,
    chunk_frames: int = 6,
    **kwargs,
) -> Iterator[torch.Tensor]:
    streamer = SoproTTSStreamer(tts, StreamConfig(chunk_frames=chunk_frames))
    return streamer.stream(
        text,
        ref_audio_path=ref_audio_path,
        ref_tokens_tq=ref_tokens_tq,
        chunk_frames=chunk_frames,
        **kwargs,
    )



from __future__ import annotations

import os
from typing import Dict, Iterator, List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from sopro.config import SoproTTSConfig
from sopro.hub import (
    download_repo,
    load_cfg_from_safetensors,
    load_state_dict_from_safetensors,
)

from .audio import save_audio
from .codec.mimi import MimiCodec
from .constants import TARGET_SR
from .nn import (
    CodebookEmbedding,
    RefXAttn,
    RMSNorm,
    SinusoidalPositionalEmbedding,
    SpeakerFiLM,
    SSMLiteBlock,
    TextEmbedding,
    TextXAttnBlock,
    Token2SV,
)
from .sampling import center_crop_tokens, repeated_tail
from .sampling import rf_ar as rf_ar_fn
from .sampling import rf_nar as rf_nar_fn
from .sampling import sample_token
from .tokenizer import TextTokenizer


class TextEncoder(nn.Module):
    def __init__(
        self,
        cfg: SoproTTSConfig,
        d_model: int,
        n_layers: int,
        tokenizer: TextTokenizer,
    ):
        super().__init__()
        self.tok = tokenizer
        self.embed = TextEmbedding(self.tok.vocab_size, d_model)
        self.layers = nn.ModuleList(
            [SSMLiteBlock(d_model, cfg.dropout, causal=False) for _ in range(n_layers)]
        )
        self.pos = SinusoidalPositionalEmbedding(d_model, max_len=cfg.max_text_len + 8)
        self.norm = RMSNorm(d_model)

    def forward(
        self, text_ids: torch.Tensor, mask: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        x = self.embed(text_ids)
        L = x.size(1)
        pos = self.pos(torch.arange(L, device=x.device))
        x = x + pos.unsqueeze(0)

        x = x * mask.unsqueeze(-1).float()
        for layer in self.layers:
            x = layer(x)
        x = self.norm(x)

        mask_f = mask.float().unsqueeze(-1)
        pooled = (x * mask_f).sum(dim=1) / (mask_f.sum(dim=1) + 1e-6)
        return x, pooled


class ARRVQ1Generator(nn.Module):
    def __init__(self, cfg: SoproTTSConfig, d_model: int, vocab: int):
        super().__init__()
        ks = cfg.ar_kernel
        dils: List[int] = []
        while len(dils) < cfg.n_layers_ar:
            dils.extend(list(cfg.ar_dilation_cycle))
        dils = dils[: cfg.n_layers_ar]

        self.dils = tuple(int(d) for d in dils)
        self.blocks = nn.ModuleList(
            [
                SSMLiteBlock(
                    d_model, cfg.dropout, causal=True, kernel_size=ks, dilation=d
                )
                for d in self.dils
            ]
        )

        self.attn_freq = int(cfg.ar_text_attn_freq)
        self.x_attns = nn.ModuleList()
        for i in range(len(self.blocks)):
            if (i + 1) % self.attn_freq == 0:
                self.x_attns.append(
                    TextXAttnBlock(d_model, heads=4, dropout=cfg.dropout)
                )
            else:
                self.x_attns.append(nn.Identity())

        self.norm = RMSNorm(d_model)
        self.head = nn.Linear(d_model, vocab)

    def forward(
        self,
        x: torch.Tensor,
        text_emb: Optional[torch.Tensor] = None,
        text_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        key_padding_mask = ~text_mask if text_mask is not None else None

        if key_padding_mask is not None:
            bad_rows = key_padding_mask.all(dim=1)
            if bad_rows.any():
                key_padding_mask = key_padding_mask.clone()
                idx = torch.nonzero(bad_rows, as_tuple=False).squeeze(1)
                key_padding_mask[idx, 0] = False
                if text_emb is not None:
                    text_emb = text_emb.clone()
                    text_emb[idx, 0, :] = 0

        h = x
        for i, lyr in enumerate(self.blocks):
            h = lyr(h)
            if not isinstance(self.x_attns[i], nn.Identity) and text_emb is not None:
                h = self.x_attns[i](h, text_emb, key_padding_mask=key_padding_mask)

        h = self.norm(h)
        return self.head(h)


class StageRefiner(nn.Module):
    def __init__(self, cfg: SoproTTSConfig, D: int, num_heads: int, codebook_size: int):
        super().__init__()
        self.blocks = nn.ModuleList(
            [SSMLiteBlock(D, cfg.dropout, causal=True) for _ in range(cfg.n_layers_nar)]
        )
        self.norm = RMSNorm(D)
        self.pre = nn.Linear(D, cfg.nar_head_dim)
        self.heads = nn.ModuleList(
            [nn.Linear(cfg.nar_head_dim, codebook_size) for _ in range(num_heads)]
        )
        self.mix = nn.Parameter(torch.ones(2, dtype=torch.float32))

    def forward_hidden(
        self, cond_bt_d: torch.Tensor, prev_bt_d: torch.Tensor
    ) -> torch.Tensor:
        w = torch.softmax(self.mix, dim=0)
        x = w[0] * cond_bt_d + w[1] * prev_bt_d
        for b in self.blocks:
            x = b(x)
        return self.norm(x)

    def forward_heads(self, h: torch.Tensor) -> List[torch.Tensor]:
        z = self.pre(h)
        return [head(z) for head in self.heads]


class StopHead(nn.Module):
    def __init__(self, D: int):
        super().__init__()
        self.proj = nn.Linear(D, 1)

    def forward(self, h: torch.Tensor) -> torch.Tensor:
        return self.proj(h).squeeze(-1)


class SoproTTSModel(nn.Module):
    def __init__(self, cfg: SoproTTSConfig, tokenizer: TextTokenizer):
        super().__init__()
        self.cfg = cfg
        D = int(cfg.d_model)

        self.text_enc = TextEncoder(cfg, D, cfg.n_layers_text, tokenizer)
        self.frame_pos = SinusoidalPositionalEmbedding(D, max_len=cfg.pos_emb_max + 2)

        self.cb_embed = CodebookEmbedding(
            cfg.num_codebooks, cfg.codebook_size, D, use_bos=True
        )
        self.rvq1_bos_id = self.cb_embed.bos_id

        self.token2sv = Token2SV(
            cfg.num_codebooks,
            cfg.codebook_size,
            d=192,
            out_dim=cfg.sv_student_dim,
            dropout=cfg.dropout,
        )
        self.spk_film = SpeakerFiLM(D, sv_dim=cfg.sv_student_dim)
        self.cond_norm = RMSNorm(D)

        self.ar = ARRVQ1Generator(cfg, D, cfg.codebook_size)
        if cfg.ar_lookback > 0:
            self.ar_hist_w = nn.Parameter(torch.zeros(cfg.ar_lookback))

        def idxs(rng: Tuple[int, int]) -> List[int]:
            lo, hi = rng
            return list(range(lo - 1, hi))

        self.stage_indices: Dict[str, List[int]] = {
            "B": idxs(cfg.stage_B),
            "C": idxs(cfg.stage_C),
            "D": idxs(cfg.stage_D),
            "E": idxs(cfg.stage_E),
        }
        self.stages = nn.ModuleDict(
            {
                s: StageRefiner(cfg, D, len(self.stage_indices[s]), cfg.codebook_size)
                for s in ["B", "C", "D", "E"]
            }
        )

        self.stop_head = StopHead(D) if cfg.use_stop_head else None

        self.ref_enc_blocks = nn.ModuleList(
            [SSMLiteBlock(D, cfg.dropout, causal=False) for _ in range(2)]
        )
        self.ref_enc_norm = RMSNorm(D)
        self.ref_xattn_stack = RefXAttn(
            D, heads=cfg.ref_attn_heads, layers=3, dropout=cfg.dropout
        )

    def rf_ar(self) -> int:
        return rf_ar_fn(self.cfg.ar_kernel, self.ar.dils)

    def rf_nar(self) -> int:
        return rf_nar_fn(self.cfg.n_layers_nar, kernel_size=7, dilation=1)

    def _pool_time(self, x: torch.Tensor, factor: int) -> torch.Tensor:
        if factor <= 1 or x.size(1) < 2 * factor:
            return x
        return F.avg_pool1d(
            x.transpose(1, 2), kernel_size=factor, stride=factor
        ).transpose(1, 2)

    def _normalize_ref_mask(
        self, ref_mask: Optional[torch.Tensor], device: torch.device
    ) -> Optional[torch.Tensor]:
        if ref_mask is None:
            return None
        mk = ref_mask.to(device).bool()
        if mk.ndim == 1:
            mk = mk.unsqueeze(0)
        return mk

    def _encode_reference_seq(self, ref_tokens: torch.Tensor) -> torch.Tensor:
        B, Tr, Q = ref_tokens.shape
        emb_sum = 0.0
        for q in range(Q):
            emb_sum = emb_sum + self.cb_embed.embed_tokens(
                ref_tokens[:, :, q], cb_index=q
            )
        x = emb_sum / float(Q)
        for b in self.ref_enc_blocks:
            x = b(x)
        return self.ref_enc_norm(x)

    def _single_pass_ref_xattn(
        self,
        cond_btd: torch.Tensor,
        ref_seq: torch.Tensor,
        ref_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        ref_seq_p = self._pool_time(ref_seq, 1)

        key_padding_mask = None
        if ref_mask is not None:
            mk_bool = ref_mask.bool()
            B, Tr = mk_bool.shape
            pooled_len = ref_seq_p.size(1)
            if pooled_len == Tr:
                key_padding_mask = ~mk_bool
            else:
                cut = pooled_len * 2
                mk2 = mk_bool[:, :cut].reshape(B, pooled_len, 2).any(dim=2)
                key_padding_mask = ~mk2

        return self.ref_xattn_stack(
            cond_btd, ref_seq_p, key_padding_mask=key_padding_mask
        )

    def _base_cond_at(
        self, t: int, txt_pool: torch.Tensor, device: torch.device
    ) -> torch.Tensor:
        pos = self.frame_pos(torch.tensor([t], device=device)).unsqueeze(0)
        return txt_pool[:, None, :] + pos

    def _ar_prev_from_seq(self, seq_1xT: torch.Tensor) -> torch.Tensor:
        K = int(self.cfg.ar_lookback)
        if K <= 0 or getattr(self, "ar_hist_w", None) is None:
            return self.cb_embed.embed_shift_by_k(seq_1xT, cb_index=0, k=1)

        ws = torch.softmax(self.ar_hist_w, dim=0)
        acc = 0.0
        k_max = min(K, int(seq_1xT.size(1)))
        for k in range(1, k_max + 1):
            acc = acc + ws[k - 1] * self.cb_embed.embed_shift_by_k(
                seq_1xT, cb_index=0, k=k
            )
        return acc

    @torch.no_grad()
    def prepare_conditioning(
        self,
        text_ids_1d: torch.Tensor,
        ref_tokens_tq: torch.Tensor,
        *,
        max_frames: int,
        device: torch.device,
        style_strength: float = 1.0,
        ref_mask: Optional[torch.Tensor] = None,
        chunk_size: Optional[int] = None,
    ) -> Dict[str, torch.Tensor]:
        self.eval()

        text_ids = text_ids_1d.to(device)
        text_mask = torch.ones_like(text_ids, dtype=torch.bool).unsqueeze(0)
        txt_seq, txt_pool = self.text_enc(text_ids.unsqueeze(0), text_mask)

        ref_btq = ref_tokens_tq.unsqueeze(0).to(device)

        sv_ref = self.token2sv(ref_btq, lengths=None)
        ref_seq = self._encode_reference_seq(ref_btq)
        ref_mask_btr = self._normalize_ref_mask(ref_mask, device)

        T = int(max_frames)
        if T <= 0:
            cond_all = torch.zeros(
                (1, 0, txt_pool.size(-1)), device=device, dtype=txt_pool.dtype
            )
        else:
            pos = self.frame_pos(torch.arange(T, device=device)).unsqueeze(0)
            base_all = txt_pool[:, None, :] + pos
            base_all = self.spk_film(base_all, sv_ref, strength=float(style_strength))

            if chunk_size is None or int(chunk_size) >= T:
                out = self._single_pass_ref_xattn(
                    base_all, ref_seq, ref_mask=ref_mask_btr
                )
                cond_all = self.cond_norm(out)
            else:
                cs = int(chunk_size)
                chunks: List[torch.Tensor] = []
                for s in range(0, T, cs):
                    e = min(T, s + cs)
                    q = base_all[:, s:e, :]
                    out = self._single_pass_ref_xattn(q, ref_seq, ref_mask=ref_mask_btr)
                    chunks.append(self.cond_norm(out))
                cond_all = torch.cat(chunks, dim=1)

        return {
            "txt_seq": txt_seq,
            "text_mask": text_mask,
            "cond_all": cond_all,
            "ref_btq": ref_btq,
            "txt_pool": txt_pool,
            "sv_ref": sv_ref,
            "ref_seq": ref_seq,
            "ref_mask": (
                ref_mask_btr
                if ref_mask_btr is not None
                else torch.empty(0, device=device, dtype=torch.bool)
            ),
            "style_strength": torch.tensor(float(style_strength), device=device),
        }

    @torch.no_grad()
    def prepare_conditioning_lazy(
        self,
        text_ids_1d: torch.Tensor,
        ref_tokens_tq: torch.Tensor,
        *,
        max_frames: int,
        device: torch.device,
        style_strength: float = 1.0,
        ref_mask: Optional[torch.Tensor] = None,
    ) -> Dict[str, torch.Tensor]:
        self.eval()

        text_ids = text_ids_1d.to(device)
        text_mask = torch.ones_like(text_ids, dtype=torch.bool).unsqueeze(0)
        txt_seq, txt_pool = self.text_enc(text_ids.unsqueeze(0), text_mask)

        ref_btq = ref_tokens_tq.unsqueeze(0).to(device)
        sv_ref = self.token2sv(ref_btq, lengths=None)
        ref_seq = self._encode_reference_seq(ref_btq)
        ref_mask_btr = self._normalize_ref_mask(ref_mask, device)

        D = int(txt_pool.size(-1))
        cond_all = torch.zeros((1, 0, D), device=device, dtype=txt_pool.dtype)

        return {
            "txt_seq": txt_seq,
            "text_mask": text_mask,
            "cond_all": cond_all,
            "ref_btq": ref_btq,
            "txt_pool": txt_pool,
            "sv_ref": sv_ref,
            "ref_seq": ref_seq,
            "ref_mask": (
                ref_mask_btr
                if ref_mask_btr is not None
                else torch.empty(0, device=device, dtype=torch.bool)
            ),
            "style_strength": torch.tensor(float(style_strength), device=device),
            "max_frames": torch.tensor(int(max_frames), device=device),
        }

    @torch.no_grad()
    def ensure_cond_upto(
        self,
        prep: Dict[str, torch.Tensor],
        t_inclusive: int,
        *,
        chunk_size: int = 64,
    ) -> None:
        if t_inclusive < 0:
            return

        cond_all = prep.get("cond_all", None)
        if cond_all is None:
            raise KeyError("prep dict missing 'cond_all'.")

        have = int(cond_all.size(1))
        need_min = int(t_inclusive) + 1
        if have >= need_min:
            return

        if "txt_pool" not in prep or "sv_ref" not in prep or "ref_seq" not in prep:
            raise RuntimeError(
                "Lazy conditioning requested but prep lacks txt_pool/sv_ref/ref_seq. "
                "Use prepare_conditioning_lazy() or prepare_conditioning(..., chunk_size=...)."
            )

        device = cond_all.device
        txt_pool = prep["txt_pool"]
        sv_ref = prep["sv_ref"]
        ref_seq = prep["ref_seq"]

        style_strength = float(
            prep.get(
                "style_strength", torch.tensor(self.cfg.style_strength, device=device)
            ).item()
        )

        ref_mask = prep.get("ref_mask", None)
        if ref_mask is not None and ref_mask.numel() == 0:
            ref_mask = None

        max_frames = prep.get("max_frames", None)
        maxT = int(max_frames.item()) if max_frames is not None else None

        cs = max(1, int(chunk_size))

        need = ((need_min + cs - 1) // cs) * cs
        if maxT is not None:
            need = min(need, maxT)

        if have >= need:
            return

        new_chunks: List[torch.Tensor] = []
        for s in range(have, need, cs):
            e = min(need, s + cs)
            pos = self.frame_pos(torch.arange(s, e, device=device)).unsqueeze(0)
            base = txt_pool[:, None, :] + pos
            base = self.spk_film(base, sv_ref, strength=style_strength)
            out = self._single_pass_ref_xattn(base, ref_seq, ref_mask=ref_mask)
            new_chunks.append(self.cond_norm(out))

        prep["cond_all"] = torch.cat([cond_all] + new_chunks, dim=1)

    def build_ar_prefix(
        self,
        ref_btq: torch.Tensor,
        device: torch.device,
        prefix_sec_fixed: Optional[float],
        use_prefix: bool,
    ) -> torch.Tensor:
        if not use_prefix or ref_btq.size(1) == 0:
            return torch.zeros(1, 0, dtype=torch.long, device=device)

        avail = int(ref_btq.size(1))
        fps = float(self.cfg.mimi_fps)

        if prefix_sec_fixed is not None and prefix_sec_fixed > 0:
            P = min(avail, int(round(prefix_sec_fixed * fps)))
        else:
            P = min(avail, max(1, int(round(self.cfg.preprompt_sec_max * fps))))

        if P <= 0:
            return torch.zeros(1, 0, dtype=torch.long, device=device)
        return ref_btq[:, :P, 0].contiguous()

    @torch.no_grad()
    def ar_stream(
        self,
        prep: Dict[str, torch.Tensor],
        *,
        max_frames: int,
        top_p: float = 0.9,
        temperature: float = 1.05,
        anti_loop: bool = True,
        loop_streak: int = 8,
        recovery_top_p: float = 0.85,
        recovery_temp: float = 1.2,
        use_prefix: bool = True,
        prefix_sec_fixed: Optional[float] = None,
        cond_chunk_size: Optional[int] = None,
        use_stop_head: Optional[bool] = None,
        stop_patience: Optional[int] = None,
        stop_threshold: Optional[float] = None,
        min_gen_frames: Optional[int] = None,
    ) -> Iterator[Tuple[int, int, Optional[float]]]:
        device = prep["cond_all"].device
        cond_all = prep["cond_all"]
        txt_seq = prep["txt_seq"]
        text_mask = prep["text_mask"]
        ref_btq = prep["ref_btq"]

        R_AR = self.rf_ar()

        stop_head = self.stop_head
        if use_stop_head is not None:
            if not bool(use_stop_head):
                stop_head = None

        eff_stop_patience = int(
            stop_patience if stop_patience is not None else self.cfg.stop_patience
        )
        eff_stop_threshold = float(
            stop_threshold if stop_threshold is not None else self.cfg.stop_threshold
        )
        eff_min_gen_frames = int(
            min_gen_frames if min_gen_frames is not None else self.cfg.min_gen_frames
        )

        A_prefix = self.build_ar_prefix(
            ref_btq, device, prefix_sec_fixed, use_prefix=use_prefix
        )
        P = int(A_prefix.size(1))

        ctx_ids = torch.zeros(
            (1, P + int(max_frames) + 1), dtype=torch.long, device=device
        )
        if P > 0:
            ctx_ids[:, :P] = A_prefix

        hist_A: List[int] = []
        loop_streak_count = 0
        stop_streak_count = 0
        last_a: Optional[int] = None

        gen_len = 0

        for t in range(int(max_frames)):
            if prep["cond_all"].size(1) < (t + 1):
                self.ensure_cond_upto(prep, t, chunk_size=int(cond_chunk_size or 64))
            cond_all = prep["cond_all"]

            L_ar = min(t + 1, R_AR)
            s_ar = t + 1 - L_ar
            cond_win_ar = cond_all[:, s_ar : t + 1, :]

            total_len = P + gen_len + 1
            A_ctx_full = ctx_ids[:, :total_len]

            prev_ctx_full = self._ar_prev_from_seq(A_ctx_full)
            prev_ctx_win = prev_ctx_full[:, -L_ar:, :]

            cur_top_p, cur_temp = top_p, temperature
            if anti_loop:
                if repeated_tail(hist_A, max_n=16):
                    cur_top_p, cur_temp = recovery_top_p, recovery_temp
                elif last_a is not None and loop_streak_count >= loop_streak:
                    cur_top_p, cur_temp = recovery_top_p, recovery_temp

            ar_logits_win = self.ar(
                cond_win_ar + prev_ctx_win, text_emb=txt_seq, text_mask=text_mask
            )
            ar_logits_t = ar_logits_win[:, -1:, :]

            rvq1_id = sample_token(
                ar_logits_t,
                history=hist_A,
                top_p=cur_top_p,
                temperature=cur_temp,
                top_k=50,
                repetition_penalty=1.1,
            )

            ctx_ids[0, P + gen_len] = int(rvq1_id)
            gen_len += 1

            hist_A.append(int(rvq1_id))
            loop_streak_count = (
                (loop_streak_count + 1)
                if (last_a is not None and rvq1_id == last_a)
                else 0
            )
            last_a = int(rvq1_id)

            p_stop: Optional[float] = None
            if stop_head is not None:
                A_now = torch.tensor([[rvq1_id]], device=device, dtype=torch.long)
                stop_inp = (
                    cond_all[:, t : t + 1, :]
                    + self.cb_embed.embed_tokens(A_now, cb_index=0).detach()
                )
                stop_logits = stop_head(stop_inp)
                p_stop = float(torch.sigmoid(stop_logits).item())

                if t + 1 >= eff_min_gen_frames and p_stop > eff_stop_threshold:
                    stop_streak_count += 1
                else:
                    stop_streak_count = 0

            yield t, int(rvq1_id), p_stop

            if stop_head is not None and stop_streak_count >= eff_stop_patience:
                break

    @torch.no_grad()
    def nar_refine(
        self, cond_seq: torch.Tensor, tokens_A_1xT: torch.Tensor
    ) -> torch.Tensor:
        preds_all: List[torch.Tensor] = [tokens_A_1xT.unsqueeze(-1)]
        prev_tokens_list: List[torch.Tensor] = [tokens_A_1xT.unsqueeze(-1)]
        prev_cb_list: List[List[int]] = [[0]]

        for stage_name in ["B", "C", "D", "E"]:
            idxs = self.stage_indices[stage_name]
            prev_tokens_cat = torch.cat(prev_tokens_list, dim=-1)
            prev_cbs_cat = sum(prev_cb_list, [])
            prev_emb_sum = self.cb_embed.sum_embed_subset(prev_tokens_cat, prev_cbs_cat)

            h = self.stages[stage_name].forward_hidden(cond_seq, prev_emb_sum)
            logits_list = self.stages[stage_name].forward_heads(h)
            preds = torch.stack([x.argmax(dim=-1) for x in logits_list], dim=-1)

            preds_all.append(preds)
            prev_tokens_list.append(preds)
            prev_cb_list.append(idxs)

        tokens_btq = torch.cat(preds_all, dim=-1)
        return tokens_btq

    @torch.no_grad()
    def generate_tokens(
        self,
        text_ids_1d: torch.Tensor,
        ref_tokens_tq: torch.Tensor,
        *,
        max_frames: int,
        device: torch.device,
        top_p: float = 0.9,
        temperature: float = 1.05,
        anti_loop: bool = True,
        use_prefix: bool = True,
        prefix_sec_fixed: Optional[float] = None,
        style_strength: float = 1.0,
        use_stop_head: Optional[bool] = None,
        stop_patience: Optional[int] = None,
        stop_threshold: Optional[float] = None,
        min_gen_frames: Optional[int] = None,
    ) -> torch.Tensor:
        prep = self.prepare_conditioning(
            text_ids_1d,
            ref_tokens_tq,
            max_frames=max_frames,
            device=device,
            style_strength=style_strength,
        )

        hist_A: List[int] = []
        for _t, rvq1, _p_stop in self.ar_stream(
            prep,
            max_frames=max_frames,
            top_p=top_p,
            temperature=temperature,
            anti_loop=anti_loop,
            use_prefix=use_prefix,
            prefix_sec_fixed=prefix_sec_fixed,
            use_stop_head=use_stop_head,
            stop_patience=stop_patience,
            stop_threshold=stop_threshold,
            min_gen_frames=min_gen_frames,
        ):
            hist_A.append(rvq1)

        T = len(hist_A)
        if T == 0:
            return torch.zeros(
                0, self.cfg.num_codebooks, dtype=torch.long, device=device
            )

        tokens_A = torch.tensor(hist_A, device=device, dtype=torch.long).unsqueeze(0)
        cond_seq = prep["cond_all"][:, :T, :]
        tokens_btq_1xTQ = self.nar_refine(cond_seq, tokens_A)
        return tokens_btq_1xTQ.squeeze(0)


class SoproTTS:
    def __init__(
        self,
        model: SoproTTSModel,
        cfg: SoproTTSConfig,
        tokenizer: TextTokenizer,
        codec: MimiCodec,
        device: str,
    ):
        self.model = model
        self.cfg = cfg
        self.tokenizer = tokenizer
        self.codec = codec
        self.device = torch.device(device)

    @classmethod
    def from_pretrained(
        cls,
        repo_id: str,
        *,
        revision: Optional[str] = None,
        cache_dir: Optional[str] = None,
        token: Optional[str] = None,
        device: Optional[str] = None,
    ) -> "SoproTTS":
        device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        dev = torch.device(device)

        local_dir = download_repo(
            repo_id, revision=revision, cache_dir=cache_dir, token=token
        )

        model_path = os.path.join(local_dir, "model.safetensors")
        if not os.path.exists(model_path):
            raise FileNotFoundError(f"Expected {model_path} in repo snapshot.")

        cfg = load_cfg_from_safetensors(model_path)

        tokenizer = TextTokenizer(model_name=local_dir)

        model = SoproTTSModel(cfg, tokenizer).to(dev).eval()
        state = load_state_dict_from_safetensors(model_path)

        model.load_state_dict(state)

        codec = MimiCodec(num_quantizers=cfg.num_codebooks, device=device)

        return cls(
            model=model, cfg=cfg, tokenizer=tokenizer, codec=codec, device=device
        )

    def encode_text(self, text: str) -> torch.Tensor:
        ids = self.tokenizer.encode(text)
        return torch.tensor(ids, dtype=torch.long, device=self.device)

    def encode_reference(
        self,
        *,
        ref_audio_path: Optional[str] = None,
        ref_tokens_tq: Optional[torch.Tensor] = None,
        ref_seconds: Optional[float] = None,
    ) -> torch.Tensor:
        if (ref_tokens_tq is None) and (ref_audio_path is None):
            raise RuntimeError(
                "SoproTTS requires a reference. Provide ref_audio_path=... or ref_tokens_tq=..."
            )
        if (ref_tokens_tq is not None) and (ref_audio_path is not None):
            raise RuntimeError(
                "Provide only one of ref_audio_path or ref_tokens_tq (not both)."
            )

        if ref_seconds is None:
            ref_seconds = float(self.cfg.ref_seconds_max)

        if ref_tokens_tq is not None:
            ref = ref_tokens_tq.to(self.device).long()
            if ref_seconds > 0:
                fps = float(self.cfg.mimi_fps)
                win = max(1, int(round(ref_seconds * fps)))
                ref = center_crop_tokens(ref, win)
            return ref

        crop_seconds = (
            ref_seconds if (ref_seconds is not None and ref_seconds > 0) else None
        )
        ref = (
            self.codec.encode_file(ref_audio_path, crop_seconds=crop_seconds)
            .to(self.device)
            .long()
        )
        return ref

    @torch.inference_mode()
    def synthesize(
        self,
        text: str,
        *,
        ref_audio_path: Optional[str] = None,
        ref_tokens_tq: Optional[torch.Tensor] = None,
        max_frames: int = 400,
        top_p: float = 0.9,
        temperature: float = 1.05,
        anti_loop: bool = True,
        use_prefix: bool = True,
        prefix_sec_fixed: Optional[float] = None,
        style_strength: Optional[float] = None,
        ref_seconds: Optional[float] = None,
        use_stop_head: Optional[bool] = None,
        stop_patience: Optional[int] = None,
        stop_threshold: Optional[float] = None,
        min_gen_frames: Optional[int] = None,
    ) -> torch.Tensor:
        text_ids = self.encode_text(text)
        ref = self.encode_reference(
            ref_audio_path=ref_audio_path,
            ref_tokens_tq=ref_tokens_tq,
            ref_seconds=ref_seconds,
        )

        tokens_tq = self.model.generate_tokens(
            text_ids,
            ref,
            max_frames=max_frames,
            device=self.device,
            top_p=top_p,
            temperature=temperature,
            anti_loop=anti_loop,
            use_prefix=use_prefix,
            prefix_sec_fixed=prefix_sec_fixed,
            style_strength=float(
                style_strength
                if style_strength is not None
                else self.cfg.style_strength
            ),
            use_stop_head=use_stop_head,
            stop_patience=stop_patience,
            stop_threshold=stop_threshold,
            min_gen_frames=min_gen_frames,
        )

        wav = self.codec.decode_full(tokens_tq)
        return wav

    def stream(self, text: str, **kwargs) -> Iterator[torch.Tensor]:
        from .streaming import stream as _stream

        return _stream(self, text, **kwargs)

    def save_wav(self, path: str, wav_1xT: torch.Tensor) -> None:
        save_audio(path, wav_1xT, sr=TARGET_SR)



from __future__ import annotations

from typing import Optional

import torch
import torch.nn as nn

from .blocks import RMSNorm


def rms(x: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    return torch.sqrt(x.pow(2).mean(dim=-1, keepdim=True) + eps)


class RefXAttnBlock(nn.Module):
    def __init__(self, d_model: int, heads: int = 2, dropout: float = 0.0):
        super().__init__()
        self.nq = RMSNorm(d_model)
        self.nkv = RMSNorm(d_model)
        self.attn = nn.MultiheadAttention(
            d_model, heads, batch_first=True, dropout=dropout
        )
        self.gate = nn.Parameter(torch.tensor(0.5))
        self.gmax = 0.35

    def forward(
        self,
        x: torch.Tensor,
        ref: torch.Tensor,
        key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        q = self.nq(x)
        kv = self.nkv(ref.float())

        with torch.autocast(device_type=q.device.type, enabled=False):
            a, _ = self.attn(
                q.float(), kv, kv, key_padding_mask=key_padding_mask, need_weights=False
            )

        a = torch.nan_to_num(a, nan=0.0, posinf=0.0, neginf=0.0)

        rms_x = rms(x.float())
        rms_a = rms(a)
        scale = rms_x / rms_a
        a = (a * scale).to(x.dtype)

        gate_eff = (self.gmax * torch.tanh(self.gate)).to(x.dtype)
        return x + gate_eff * a


class RefXAttn(nn.Module):
    def __init__(
        self, d_model: int, heads: int = 2, layers: int = 3, dropout: float = 0.0
    ):
        super().__init__()
        self.blocks = nn.ModuleList(
            [RefXAttnBlock(d_model, heads, dropout) for _ in range(layers)]
        )

    def forward(
        self,
        x: torch.Tensor,
        ref: torch.Tensor,
        key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        for blk in self.blocks:
            x = blk(x, ref, key_padding_mask)
        return x


class TextXAttnBlock(nn.Module):
    def __init__(self, d_model: int, heads: int = 4, dropout: float = 0.0):
        super().__init__()
        self.nq = RMSNorm(d_model)
        self.nkv = RMSNorm(d_model)
        self.attn = nn.MultiheadAttention(
            d_model, num_heads=heads, dropout=dropout, batch_first=True
        )
        self.gate = nn.Parameter(torch.tensor(0.0))

    def forward(
        self,
        x: torch.Tensor,
        context: torch.Tensor,
        key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        q = self.nq(x)
        kv = self.nkv(context)
        with torch.autocast(device_type=q.device.type, enabled=False):
            out, _ = self.attn(
                q.float(),
                kv.float(),
                kv.float(),
                key_padding_mask=key_padding_mask,
                need_weights=False,
            )
        out = torch.nan_to_num(out, nan=0.0, posinf=0.0, neginf=0.0).to(x.dtype)
        return x + torch.tanh(self.gate) * out



from __future__ import annotations

from typing import List

import torch


def center_crop_tokens(ref_tq: torch.Tensor, win_frames: int) -> torch.Tensor:
    T = int(ref_tq.size(0))
    if T <= win_frames:
        return ref_tq
    s = (T - win_frames) // 2
    return ref_tq[s : s + win_frames]


def repeated_tail(hist: List[int], max_n: int = 16) -> bool:
    L = len(hist)
    for n in range(3, min(max_n, L // 2) + 1):
        if hist[-n:] == hist[-2 * n : -n]:
            return True
    return False


def sample_token(
    logits_1x1v: torch.Tensor,
    history: List[int],
    top_p: float = 0.9,
    top_k: int = 0,
    temperature: float = 1.0,
    repetition_penalty: float = 1.0,
    eps: float = 1e-12,
) -> int:
    x = logits_1x1v

    x = torch.nan_to_num(x, nan=-1e9, posinf=1e9, neginf=-1e9)

    if temperature and temperature != 1.0:
        x = x / float(temperature)

    if repetition_penalty != 1.0 and len(history) > 0:
        context = history[-50:]
        ids = torch.tensor(list(set(context)), device=x.device, dtype=torch.long)
        if ids.numel() > 0:
            vals = x[0, 0, ids]
            neg = vals < 0
            vals = torch.where(
                neg, vals * repetition_penalty, vals / repetition_penalty
            )
            x = x.clone()
            x[0, 0, ids] = vals

    probs = torch.softmax(x, dim=-1).view(1, -1)
    probs = torch.nan_to_num(probs, nan=0.0, posinf=0.0, neginf=0.0)

    V = int(probs.size(-1))
    if top_k and top_k > 0:
        k = min(int(top_k), V)
        val, idx = torch.topk(probs, k, dim=-1)
        newp = torch.zeros_like(probs)
        newp.scatter_(1, idx, val)
        probs = newp

        s = probs.sum(dim=-1, keepdim=True)
        if float(s.item()) <= eps:
            return int(torch.argmax(x[0, 0]).item())
        probs = probs / s

    if top_p is not None and top_p < 1.0:
        sorted_probs, sorted_idx = torch.sort(probs, descending=True, dim=-1)
        cum = torch.cumsum(sorted_probs, dim=-1)

        remove = cum > float(top_p)
        remove[..., 1:] = remove[..., :-1].clone()
        remove[..., 0] = False

        sorted_probs = sorted_probs.masked_fill(remove, 0.0)

        s = sorted_probs.sum(dim=-1, keepdim=True)
        if float(s.item()) <= eps:
            return int(torch.argmax(x[0, 0]).item())
        sorted_probs = sorted_probs / s

        j = torch.multinomial(sorted_probs, 1).item()
        token = int(sorted_idx[0, j].item())

        return token

    s = probs.sum(dim=-1, keepdim=True)
    if float(s.item()) <= eps:
        return int(torch.argmax(x[0, 0]).item())
    probs = probs / s

    return int(torch.multinomial(probs, 1).item())


def rf_ar(ar_kernel: int, dilations: Tuple[int, ...]) -> int:
    return 1 + (ar_kernel - 1) * int(sum(dilations))


def rf_nar(n_layers_nar: int, kernel_size: int = 7, dilation: int = 1) -> int:
    return 1 + (kernel_size - 1) * int(n_layers_nar) * int(dilation)



from __future__ import annotations

import os
from typing import Optional, Tuple

import torch
import torch.nn.functional as F

from .constants import TARGET_SR

try:
    import torchaudio
    import torchaudio.functional as AF
except Exception:
    torchaudio = None
    AF = None

try:
    import soundfile as sf
except Exception:
    sf = None


def device_str() -> str:
    if torch.cuda.is_available():
        return "cuda"
    return "cpu"


def trim_silence_energy(
    wav: torch.Tensor,
    sr: int,
    frame_ms: float = 25.0,
    hop_ms: float = 10.0,
    thresh_db_floor: float = -40.0,
    prepad_ms: float = 30.0,
    postpad_ms: float = 30.0,
    min_keep_sec: float = 0.5,
) -> torch.Tensor:
    orig_1d = wav.ndim == 1
    if orig_1d:
        wav = wav.unsqueeze(0)

    C, T = wav.shape
    if T == 0:
        return wav.squeeze(0) if orig_1d else wav
    if T < int(sr * 0.1):
        return wav.squeeze(0) if orig_1d else wav

    frame_len = max(1, int(sr * frame_ms / 1000.0))
    hop = max(1, int(sr * hop_ms / 1000.0))
    if T < frame_len:
        return wav.squeeze(0) if orig_1d else wav

    mono = wav.mean(dim=0, keepdim=True)
    frames = mono.unfold(-1, frame_len, hop)
    energy = frames.pow(2).mean(dim=-1).squeeze(0)

    eps = 1e-10
    energy_db = 10.0 * torch.log10(energy + eps)
    max_db = float(energy_db.max().item())

    rel_thresh = max_db + thresh_db_floor
    thresh_db = max(rel_thresh, thresh_db_floor)

    voiced = energy_db > thresh_db
    idx = torch.nonzero(voiced, as_tuple=False)
    if idx.numel() == 0:
        return wav.squeeze(0) if orig_1d else wav

    first_frame = int(idx[0, 0].item())
    last_frame = int(idx[-1, 0].item())

    prepad_samples = int(sr * prepad_ms / 1000.0)
    postpad_samples = int(sr * postpad_ms / 1000.0)

    start = max(0, first_frame * hop - prepad_samples)

    end = min(T, last_frame * hop + frame_len + postpad_samples)

    min_keep = int(min_keep_sec * sr)
    if end <= start or (end - start) < min_keep:
        return wav.squeeze(0) if orig_1d else wav

    out = wav[:, start:end]
    return out.squeeze(0) if orig_1d else out


def load_audio_file(path: str) -> Tuple[torch.Tensor, int]:
    if sf is not None:
        wav_np, sr = sf.read(path, dtype="float32", always_2d=True)
        wav = torch.from_numpy(wav_np).transpose(0, 1)
    elif torchaudio is not None:
        wav, sr = torchaudio.load(path)
        if wav.dtype != torch.float32:
            if wav.dtype == torch.int16:
                wav = wav.float() / (2**15)
            else:
                wav = wav.float()
    else:
        raise RuntimeError("Install 'soundfile' or 'torchaudio' to read audio.")

    if wav.size(0) > 1:
        wav = wav.mean(dim=0, keepdim=True)
    return wav, sr


def resample(
    wav: torch.Tensor, sr_in: int, sr_out: int, device: Optional[str] = None
) -> torch.Tensor:
    device = device or device_str()
    wav = wav.to(device)
    if sr_in == sr_out:
        return wav
    if sr_in != sr_out and AF is None:
        raise RuntimeError("Resampling requires torchaudio. pip install torchaudio")
    return AF.resample(wav, sr_in, sr_out)


def save_audio(path: str, wav: torch.Tensor, sr: int = TARGET_SR) -> None:
    os.makedirs(os.path.dirname(os.path.abspath(path)) or ".", exist_ok=True)

    wav = wav.detach().cpu()

    if wav.ndim == 1:
        wav = wav.unsqueeze(0)
    elif wav.ndim == 2:
        pass
    elif wav.ndim == 3:
        wav = wav[0]
    else:
        raise ValueError(f"Expected wav with 1-3 dims, got shape {tuple(wav.shape)}")

    if wav.size(0) > 1:
        wav = wav.mean(dim=0, keepdim=True)

    if sf is not None:
        sf.write(path, wav[0].numpy(), sr)
        return

    if torchaudio is not None:
        torchaudio.save(path, wav, sample_rate=sr)
        return

    raise RuntimeError("Install 'soundfile' or 'torchaudio' to write audio.")


def center_crop_audio(wav: torch.Tensor, win_samples: int) -> torch.Tensor:
    if win_samples <= 0:
        return wav
    T = int(wav.shape[-1])
    if T <= win_samples:
        return wav
    s = (T - win_samples) // 2
    return wav[..., s : s + win_samples]



from __future__ import annotations

from typing import List

try:
    from transformers import AutoTokenizer
    from transformers import logging as hf_logging

    hf_logging.set_verbosity_error()
except Exception:
    hf_logging = None
    AutoTokenizer = None


class TextTokenizer:
    def __init__(self, model_name: str, add_bos_eos: bool = True):
        if AutoTokenizer is None:
            raise ImportError("pip install transformers sentencepiece")
        self.tok = AutoTokenizer.from_pretrained(model_name, use_fast=True)
        self.add_bos_eos = add_bos_eos

        if self.tok.pad_token_id is None:
            self.tok.add_special_tokens({"pad_token": "<|pad|>"})

        self.pad_id = int(self.tok.pad_token_id)
        self.bos_id = (
            int(self.tok.bos_token_id) if self.tok.bos_token_id is not None else None
        )
        self.eos_id = (
            int(self.tok.eos_token_id) if self.tok.eos_token_id is not None else None
        )
        self.vocab_size = int(self.tok.vocab_size + len(self.tok.get_added_vocab()))

    def encode(self, text: str) -> List[int]:
        ids = self.tok.encode(text, add_special_tokens=False)
        if self.add_bos_eos and self.bos_id is not None and self.eos_id is not None:
            ids = [self.bos_id] + ids + [self.eos_id]
        return ids



from .blocks import GLU, AttentiveStatsPool, DepthwiseConv1d, RMSNorm, SSMLiteBlock
from .embeddings import CodebookEmbedding, SinusoidalPositionalEmbedding, TextEmbedding
from .speaker import SpeakerFiLM, Token2SV
from .xattn import RefXAttn, RefXAttnBlock, TextXAttnBlock

__all__ = [
    "GLU",
    "RMSNorm",
    "DepthwiseConv1d",
    "SSMLiteBlock",
    "AttentiveStatsPool",
    "SinusoidalPositionalEmbedding",
    "TextEmbedding",
    "CodebookEmbedding",
    "Token2SV",
    "SpeakerFiLM",
    "RefXAttn",
    "RefXAttnBlock",
    "TextXAttnBlock",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional, Tuple

import torch

from ..audio import center_crop_audio, load_audio_file, resample, trim_silence_energy
from ..constants import DEFAULT_MIMI_ID, TARGET_SR

try:
    from transformers import MimiConfig, MimiModel
except Exception:
    MimiModel = None
    MimiConfig = None


class MimiCodec:
    def __init__(
        self, num_quantizers: int, device: str = "cuda", model_id: str = DEFAULT_MIMI_ID
    ):
        if MimiModel is None or MimiConfig is None:
            raise RuntimeError(
                "MimiModel missing. Install a recent transformers version."
            )

        self.device = torch.device(device)
        cfg = MimiConfig.from_pretrained(model_id, num_quantizers=int(num_quantizers))
        self.model = (
            MimiModel.from_pretrained(model_id, config=cfg).to(self.device).eval()
        )

    @property
    def codebook_size(self) -> int:
        return int(getattr(self.model.config, "codebook_size", 2048))

    @property
    def num_quantizers(self) -> int:
        return int(getattr(self.model.config, "num_quantizers", 32))

    @torch.no_grad()
    def encode_file(
        self, wav_path: str, *, crop_seconds: Optional[float] = None
    ) -> torch.Tensor:
        wav, sr = load_audio_file(wav_path)

        wav = trim_silence_energy(wav, sr)

        cfg = self.model.config
        sr_target = int(getattr(cfg, "sampling_rate", TARGET_SR))
        wav = resample(wav, sr, sr_target, device=str(self.device))

        if crop_seconds is not None and crop_seconds > 0:
            fps = float(getattr(cfg, "frame_rate", 12.5))
            hop = int(round(sr_target / fps))
            win_frames = max(1, int(round(crop_seconds * fps)))
            win_samples = win_frames * hop
            wav = center_crop_audio(wav, win_samples)

        wav = wav.unsqueeze(0)
        out = self.model.encode(wav, return_dict=True)
        codes_bqt = out.audio_codes
        return codes_bqt[0].permute(1, 0).contiguous()

    @torch.no_grad()
    def decode_full(self, codes_tq: torch.Tensor) -> torch.Tensor:
        audio_codes = codes_tq.to(self.device).permute(1, 0).unsqueeze(0).contiguous()
        out = self.model.decode(audio_codes=audio_codes, return_dict=True)
        wav = out.audio_values
        if wav.ndim == 1:
            wav = wav.unsqueeze(0)
        return wav


@dataclass
class MimiDecodeState:
    decoder_past_key_values: Optional[object] = None
    frames_seen: int = 0
    samples_emitted: int = 0
    tail_codes_tq: Optional[torch.Tensor] = None


class MimiStreamDecoder:
    def __init__(self, codec: MimiCodec):
        self.codec = codec
        self.codec.model.config.use_cache = True

    def drop_cache_tail(self, pkv: Any, n: int):
        if pkv is None or n <= 0:
            return pkv

        if hasattr(pkv, "to_legacy_cache") and hasattr(type(pkv), "from_legacy_cache"):
            legacy = pkv.to_legacy_cache()

            trimmed = tuple(
                (
                    k[..., : max(0, k.shape[-2] - n), :].contiguous(),
                    v[..., : max(0, v.shape[-2] - n), :].contiguous(),
                )
                for (k, v) in legacy
            )
            return type(pkv).from_legacy_cache(trimmed)

        if isinstance(pkv, tuple):
            return tuple(
                (
                    k[..., : max(0, k.shape[-2] - n), :].contiguous(),
                    v[..., : max(0, v.shape[-2] - n), :].contiguous(),
                )
                for (k, v) in pkv
            )

        return pkv

    @torch.inference_mode()
    def decode_step(
        self,
        codes_chunk_tq: torch.Tensor,
        state: Optional[MimiDecodeState] = None,
        *,
        overlap_frames: int = 2,
    ) -> Tuple[torch.Tensor, MimiDecodeState]:
        if state is None:
            state = MimiDecodeState()

        cfg = self.codec.model.config
        sr = int(getattr(cfg, "sampling_rate", 24000))
        fps = float(getattr(cfg, "frame_rate", 12.5))
        hop = int(round(sr / fps))

        n_new = int(codes_chunk_tq.size(0))
        if n_new == 0:
            return torch.zeros(1, 0, device=self.codec.device), state

        tail = state.tail_codes_tq
        ov = 0
        if overlap_frames > 0 and tail is not None and tail.numel() > 0:
            ov = min(int(overlap_frames), int(tail.size(0)))
            tail = tail[-ov:]
            codes_in_tq = torch.cat([tail, codes_chunk_tq], dim=0)
        else:
            codes_in_tq = codes_chunk_tq

        pkv = state.decoder_past_key_values
        if ov > 0 and pkv is not None:
            pkv = self.drop_cache_tail(pkv, ov)

        audio_codes = (
            codes_in_tq.to(self.codec.device).permute(1, 0).unsqueeze(0).contiguous()
        )

        out = self.codec.model.decode(
            audio_codes=audio_codes,
            decoder_past_key_values=pkv,
            return_dict=True,
        )

        wav = out.audio_values
        if wav.ndim == 1:
            wav = wav.unsqueeze(0)
        else:
            wav = wav.reshape(1, -1)

        expected_total = int((ov + n_new) * hop)
        if wav.size(1) >= expected_total:
            wav = wav[:, :expected_total]

        ov_samp = min(int(ov * hop), int(wav.size(1)))
        wav_new = wav[:, ov_samp:]

        state.decoder_past_key_values = getattr(out, "decoder_past_key_values", None)
        state.frames_seen += n_new
        state.samples_emitted += int(wav_new.size(1))

        if overlap_frames > 0:
            keep = min(int(overlap_frames), int(codes_in_tq.size(0)))
            state.tail_codes_tq = codes_in_tq[-keep:].detach()
        else:
            state.tail_codes_tq = None

        return wav_new, state



from dataclasses import dataclass
from typing import Tuple

from sopro.constants import TARGET_SR


@dataclass
class SoproTTSConfig:
    num_codebooks: int = 32
    codebook_size: int = 2048
    mimi_fps: float = 12.5
    max_frames: int = 400
    audio_sr: int = TARGET_SR

    d_model: int = 384
    n_layers_text: int = 4
    n_layers_ar: int = 6
    n_layers_nar: int = 6
    dropout: float = 0.05

    pos_emb_max: int = 4096
    max_text_len: int = 2048

    nar_head_dim: int = 256

    use_stop_head: bool = True
    stop_threshold: float = 0.8
    stop_patience: int = 5
    min_gen_frames: int = 12

    stage_B: Tuple[int, int] = (2, 4)
    stage_C: Tuple[int, int] = (5, 8)
    stage_D: Tuple[int, int] = (9, 16)
    stage_E: Tuple[int, int] = (17, 32)

    ar_lookback: int = 4
    ar_kernel: int = 13
    ar_dilation_cycle: Tuple[int, ...] = (1, 2, 4, 1)

    ar_text_attn_freq: int = 2

    ref_attn_heads: int = 2
    ref_seconds_max: float = 12.0

    preprompt_sec_max: float = 4.0

    sv_student_dim: int = 192
    style_strength: float = 1.0



from __future__ import annotations

import math
from typing import List, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


class SinusoidalPositionalEmbedding(nn.Module):
    def __init__(self, d_model: int, max_len: int = 10000):
        super().__init__()
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2, dtype=torch.float32)
            * (-math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe, persistent=False)

    def forward(self, positions: torch.Tensor) -> torch.Tensor:
        return self.pe.index_select(0, positions.long())


class TextEmbedding(nn.Module):
    def __init__(self, vocab_size: int, d_model: int):
        super().__init__()
        self.emb = nn.Embedding(vocab_size, d_model)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.emb(x)


class CodebookEmbedding(nn.Module):
    def __init__(
        self, num_codebooks: int, vocab_size: int, d_model: int, use_bos: bool = True
    ):
        super().__init__()
        self.Q = int(num_codebooks)
        self.V = int(vocab_size)
        self.D = int(d_model)
        self.use_bos = bool(use_bos)

        table_size = self.Q * self.V + (1 if self.use_bos else 0)
        self.emb = nn.Embedding(table_size, d_model)
        self.bos_id = (self.Q * self.V) if self.use_bos else None

    def _indices_for(self, tokens: torch.Tensor, cb_index: int) -> torch.Tensor:
        return cb_index * self.V + tokens

    def embed_tokens(self, tokens: torch.Tensor, cb_index: int) -> torch.Tensor:
        return self.emb(self._indices_for(tokens, cb_index))

    def embed_shift_by_k(
        self, tokens: torch.Tensor, cb_index: int, k: int
    ) -> torch.Tensor:
        idx = self._indices_for(tokens, cb_index)
        B, T = idx.shape
        if (not self.use_bos) or (self.bos_id is None) or k <= 0:
            pad_tok = idx[:, :1]
        else:
            pad_tok = torch.full(
                (B, 1), self.bos_id, dtype=torch.long, device=idx.device
            )

        if k >= T:
            idx_shift = pad_tok.expand(-1, T)
        else:
            pad = pad_tok.expand(-1, k)
            idx_shift = torch.cat([pad, idx[:, :-k]], dim=1)

        return self.emb(idx_shift)

    def sum_embed_subset(
        self,
        tokens_subset: Optional[torch.Tensor],
        cb_indices: Optional[List[int]],
        keep_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if tokens_subset is None or cb_indices is None or len(cb_indices) == 0:
            return 0.0

        B, T, K = tokens_subset.shape
        idx_list = []
        for k, cb in enumerate(cb_indices):
            idx_list.append(self._indices_for(tokens_subset[..., k], cb).unsqueeze(2))
        idx = torch.cat(idx_list, dim=2)
        emb = self.emb(idx)

        if keep_mask is not None:
            emb = emb * keep_mask.unsqueeze(-1).to(emb.dtype)

        return emb.sum(dim=2)



from __future__ import annotations

import json
import struct
from typing import Any, Dict, Optional

import torch
from huggingface_hub import snapshot_download
from safetensors.torch import load_file

from sopro.config import SoproTTSConfig


def download_repo(
    repo_id: str,
    *,
    revision: Optional[str] = None,
    cache_dir: Optional[str] = None,
    token: Optional[str] = None,
) -> str:
    return snapshot_download(
        repo_id=repo_id,
        revision=revision,
        cache_dir=cache_dir,
        token=token,
    )


def _read_safetensors_metadata(path: str) -> Dict[str, str]:
    with open(path, "rb") as f:
        header_len = struct.unpack("<Q", f.read(8))[0]
        header = json.loads(f.read(header_len).decode("utf-8"))
    meta = header.get("__metadata__", {}) or {}
    return {str(k): str(v) for k, v in meta.items()}


def load_cfg_from_safetensors(path: str) -> SoproTTSConfig:
    meta = _read_safetensors_metadata(path)
    if "cfg" not in meta:
        raise RuntimeError(f"No 'cfg' metadata found in {path}.")

    cfg_dict = json.loads(meta["cfg"])
    init: Dict[str, Any] = {}
    for k in SoproTTSConfig.__annotations__.keys():
        if k in cfg_dict:
            init[k] = cfg_dict[k]

    cfg = SoproTTSConfig(**init)
    return cfg


def load_state_dict_from_safetensors(path: str) -> Dict[str, torch.Tensor]:
    return load_file(path)
